@@ -17,6 +17,8 @@ pub use output::*;
 pub mod query;
 pub use query::*;
 
+const USER_AGENT: &str = concat!("bodhi-cli v", env!("CARGO_PKG_VERSION"));
+
 /// This function prompts the user for their FAS password.
 fn read_password() -> String {
     rpassword::prompt_password_stdout("FAS Password: ").expect("Failed to read from console.")
@@ -47,7 +49,7 @@ fn get_store_password(clear: bool) -> Result<String, String> {
         if let Err(error) = collection.create_item(
             "bodhi-cli",
             attributes.clone(),
-            &password.as_bytes(),
+            password.as_bytes(),
             replace,
             "password",
         ) {
@@ -103,38 +105,27 @@ fn get_store_password(clear: bool) -> Result<String, String> {
 }
 
 #[allow(clippy::cognitive_complexity)]
-fn main() -> Result<(), String> {
+#[tokio::main]
+async fn main() -> Result<(), String> {
     let args: BaseCommand = BaseCommand::from_args();
+    let authenticated = args.authenticated();
 
-    let authenticated = match &args.subcommand {
-        BodhiCommand::Comment { .. } => true,
-        BodhiCommand::ComposeInfo { .. } => false,
-        BodhiCommand::ComposeList { .. } => false,
-        BodhiCommand::CreateOverride { .. } => true,
-        BodhiCommand::CreateUpdate { .. } => true,
-        BodhiCommand::EditOverride { .. } => true,
-        BodhiCommand::EditUpdate { .. } => true,
-        BodhiCommand::ExpireOverride { .. } => true,
-        BodhiCommand::QueryOverrides { .. } => false,
-        BodhiCommand::QueryUpdates { .. } => false,
-        BodhiCommand::ReleaseInfo { .. } => false,
-        BodhiCommand::ReleaseList { .. } => false,
-        BodhiCommand::UpdateRequest { .. } => true,
-        BodhiCommand::WaiveTests { .. } => true,
-    };
-
-    let config = get_config()?;
+    let config = get_config().await?;
 
     let mut builder = match (&args.staging, &args.bodhi_url, &args.login_url) {
-        (false, None, None) => BodhiServiceBuilder::default(),
-        (true, None, None) => BodhiServiceBuilder::staging(),
-        (false, Some(url), Some(login_url)) => BodhiServiceBuilder::custom(url.to_owned(), login_url.to_owned()),
+        (false, None, None) => BodhiClientBuilder::default(),
+        (true, None, None) => BodhiClientBuilder::staging(),
+        (false, Some(url), Some(login_url)) => BodhiClientBuilder::custom(url.to_owned(), login_url.to_owned()),
         _ => unreachable!(),
     };
 
+    builder = builder.user_agent(USER_AGENT);
+
     let bodhi = if authenticated {
-        println!("Authenticating with bodhi ...");
-        println!("Username: {}", &config.fas.username);
+        if args.verbose {
+            eprintln!("Authenticating with bodhi ...");
+            eprintln!("Username: {}", &config.fas.username);
+        }
 
         let password = if !args.no_store_password {
             get_store_password(args.ignore_keyring)?
@@ -143,27 +134,17 @@ fn main() -> Result<(), String> {
         };
 
         builder = builder.authentication(&config.fas.username, &password);
-
-        match builder.build() {
-            Ok(bodhi) => bodhi,
-            Err(error) => return Err(error.to_string()),
-        }
+        builder.build().await.map_err(|error| error.to_string())?
     } else {
-        match builder.build() {
-            Ok(bodhi) => bodhi,
-            Err(error) => return Err(error.to_string()),
-        }
+        builder.build().await.map_err(|error| error.to_string())?
     };
 
     match args.subcommand {
         BodhiCommand::Comment { update, text, karma } => {
-            let update: Update = match bodhi.query(UpdateIDQuery::new(&update)) {
-                Ok(update) => match update {
-                    Some(update) => update,
-                    None => return Err(String::from("Update not found.")),
-                },
-                Err(error) => return Err(error.to_string()),
-            };
+            let update: Update = bodhi
+                .request(&UpdateIDQuery::new(&update))
+                .await
+                .map_err(|error| error.to_string())?;
 
             let mut commenter = update.comment().text(&text);
 
@@ -171,38 +152,33 @@ fn main() -> Result<(), String> {
                 commenter = commenter.karma(karma);
             }
 
-            match bodhi.create(&commenter) {
-                Ok(_) => {
-                    println!("Comment created.");
-                    Ok(())
-                },
-                Err(error) => Err(error.to_string()),
-            }
+            let comment: NewComment = bodhi.request(&commenter).await.map_err(|error| error.to_string())?;
+
+            println!("Comment created.");
+            print_server_msgs(&comment.caveats);
+            println!("{}", &comment.comment);
+
+            Ok(())
         },
         BodhiCommand::ComposeInfo {
             release,
             request,
             format,
         } => {
-            let result: Option<Compose> = match bodhi.query(ComposeReleaseRequestQuery::new(release, request)) {
-                Ok(compose) => compose,
-                Err(error) => return Err(error.to_string()),
-            };
+            let result: Compose = bodhi
+                .request(&ComposeReleaseRequestQuery::new(&release, request))
+                .await
+                .map_err(|error| error.to_string())?;
 
-            pretty_output(
-                result.as_ref(),
-                &format!("{}/{}", release, request),
-                "No running compose found for",
-                format.unwrap_or(Format::Plain),
-            )?;
+            pretty_output(&result, format.unwrap_or(Format::Plain))?;
 
             Ok(())
         },
         BodhiCommand::ComposeList { format } => {
-            let result: Vec<Compose> = match bodhi.query(ComposeQuery::new()) {
-                Ok(composes) => composes,
-                Err(error) => return Err(error.to_string()),
-            };
+            let result: Vec<Compose> = bodhi
+                .request(&ComposeQuery::new())
+                .await
+                .map_err(|error| error.to_string())?;
 
             pretty_outputs(&result, format.unwrap_or(Format::Plain))?;
 
@@ -212,17 +188,46 @@ fn main() -> Result<(), String> {
             let current_date = chrono::Utc::now();
             let expiration_date = (current_date + chrono::Duration::days(duration as i64)).into();
 
-            let builder = OverrideBuilder::new(&nvr, &notes, &expiration_date);
+            let creator = OverrideCreator::new(&nvr, &notes, &expiration_date);
 
-            let result: NewOverride = match bodhi.create(&builder) {
-                Ok(o) => o,
-                Err(error) => return Err(error.to_string()),
+            return match bodhi.request(&creator).await {
+                Ok(result) => {
+                    println!("Successfully created override for: {}", &result.over_ride.nvr);
+                    print_server_msgs(&result.caveats);
+                    Ok(())
+                },
+                Err(error) => {
+                    println!("Failed to create override for {}, aborting.", &nvr);
+                    Err(error.to_string())
+                },
             };
+        },
+        BodhiCommand::CreateUpdateOverride { alias, duration, notes } => {
+            let update = query_update(&bodhi, &alias).await?;
 
-            print_server_msgs(&result.caveats);
-            println!("{}", result.over_ride);
+            let current_date = chrono::Utc::now();
+            let expiration_date = (current_date + chrono::Duration::days(duration as i64)).into();
 
-            Ok(())
+            let mut result = Ok(());
+
+            for build in &update.builds {
+                let creator = OverrideCreator::new(&build.nvr, &notes, &expiration_date);
+
+                match bodhi.request(&creator).await {
+                    Ok(result) => {
+                        println!(" - successfully created override for: {}", &result.over_ride.nvr);
+                        print_server_msgs(&result.caveats);
+                        continue;
+                    },
+                    Err(error) => {
+                        println!(" - failed to create override for {}, aborting.", &build.nvr);
+                        result = Err(format!("{}", error));
+                        break;
+                    },
+                }
+            }
+
+            result
         },
         BodhiCommand::CreateUpdate {
             autokarma,
@@ -249,8 +254,8 @@ fn main() -> Result<(), String> {
 
             let mut builder = match (&builds, &from_tag) {
                 (Some(_), Some(_)) => unreachable!(),
-                (Some(builds), None) => UpdateBuilder::from_builds(&builds, &notes),
-                (None, Some(tag)) => UpdateBuilder::from_tag(&tag, &notes),
+                (Some(builds), None) => UpdateCreator::from_builds(builds, &notes),
+                (None, Some(tag)) => UpdateCreator::from_tag(tag, &notes),
                 (None, None) => return Err(String::from("Neither builds nor koji tag specified.")),
             };
 
@@ -262,17 +267,15 @@ fn main() -> Result<(), String> {
                 builder = builder.autotime(autotime);
             };
 
-            if let Some(bugs) = bugs {
-                for bug in bugs {
-                    builder = builder.bugs(bug);
-                }
+            if let Some(bugs) = &bugs {
+                builder = builder.bugs(bugs);
             };
 
             if let Some(close_bugs) = close_bugs {
                 builder = builder.close_bugs(close_bugs);
             };
 
-            if let Some(display_name) = display_name {
+            if let Some(display_name) = &display_name {
                 builder = builder.display_name(display_name);
             };
 
@@ -284,9 +287,10 @@ fn main() -> Result<(), String> {
                 builder = builder.require_testcases(require_testcases);
             };
 
-            if let Some(requirements) = &requirements {
+            let requirements_string = requirements.map(|some| some.join(","));
+            if let Some(requirements) = &requirements_string {
                 if !requirements.is_empty() {
-                    builder = builder.requirements(requirements.join(","));
+                    builder = builder.requirements(requirements);
                 };
             };
 
@@ -314,11 +318,9 @@ fn main() -> Result<(), String> {
                 builder = builder.update_type(update_type);
             };
 
-            let result: NewUpdate = match bodhi.create(&builder) {
-                Ok(value) => value,
-                Err(error) => return Err(error.to_string()),
-            };
+            let result: NewUpdate = bodhi.request(&builder).await.map_err(|error| error.to_string())?;
 
+            println!("Update created.");
             print_server_msgs(&result.caveats);
             println!("{}", result.update);
 
@@ -328,16 +330,14 @@ fn main() -> Result<(), String> {
             let current_date = chrono::Utc::now();
             let expiration_date = (current_date + chrono::Duration::days(duration as i64)).into();
 
-            let over_ride = query_override(&bodhi, &nvr)?;
+            let over_ride = query_override(&bodhi, &nvr).await?;
             let editor = OverrideEditor::from_override(&over_ride)
                 .expiration_date(&expiration_date)
                 .notes(&notes);
 
-            let result: EditedOverride = match bodhi.edit(&editor) {
-                Ok(value) => value,
-                Err(error) => return Err(error.to_string()),
-            };
+            let result: EditedOverride = bodhi.request(&editor).await.map_err(|error| error.to_string())?;
 
+            println!("Override edited.");
             print_server_msgs(&result.caveats);
             println!("{}", result.over_ride);
 
@@ -362,7 +362,7 @@ fn main() -> Result<(), String> {
             unstable_karma,
             update_type,
         } => {
-            let update = query_update(&bodhi, &alias)?;
+            let update = query_update(&bodhi, &alias).await?;
             let mut editor = UpdateEditor::from_update(&update);
 
             if let Some(add_bugs) = add_bugs {
@@ -411,7 +411,7 @@ fn main() -> Result<(), String> {
 
             let requirements = requirements.as_ref().map(|reqs| reqs.join(","));
             if let Some(requirements) = &requirements {
-                editor = editor.requirements(&requirements);
+                editor = editor.requirements(requirements);
             }
 
             if let Some(severity) = severity {
@@ -438,25 +438,21 @@ fn main() -> Result<(), String> {
                 editor = editor.update_type(update_type);
             }
 
-            let result: EditedUpdate = match bodhi.edit(&editor) {
-                Ok(value) => value,
-                Err(error) => return Err(error.to_string()),
-            };
+            let result: EditedUpdate = bodhi.request(&editor).await.map_err(|error| error.to_string())?;
 
+            println!("Update edited.");
             print_server_msgs(&result.caveats);
             println!("{}", result.update);
 
             Ok(())
         },
         BodhiCommand::ExpireOverride { nvr } => {
-            let over_ride = query_override(&bodhi, &nvr)?;
+            let over_ride = query_override(&bodhi, &nvr).await?;
             let editor = OverrideEditor::from_override(&over_ride).expired(true);
 
-            let result: EditedOverride = match bodhi.edit(&editor) {
-                Ok(value) => value,
-                Err(error) => return Err(error.to_string()),
-            };
+            let result: EditedOverride = bodhi.request(&editor).await.map_err(|error| error.to_string())?;
 
+            println!("Override expired.");
             print_server_msgs(&result.caveats);
             println!("{}", result.over_ride);
 
@@ -472,11 +468,14 @@ fn main() -> Result<(), String> {
         } => {
             let format = format.unwrap_or(Format::Plain);
 
+            let build_refs: Option<Vec<&str>> = builds.as_ref().map(|bs| bs.iter().map(|b| b.as_str()).collect());
+            let user_refs: Option<Vec<&str>> = users.as_ref().map(|us| us.iter().map(|u| u.as_str()).collect());
+
             let mut query = OverrideQuery::new();
             let mut long_running = true;
 
-            if let Some(builds) = &builds {
-                query = query.builds(builds.iter().map(|s| s.as_str()).collect());
+            if let Some(build_refs) = &build_refs {
+                query = query.builds(build_refs);
                 long_running = false;
             };
 
@@ -485,13 +484,13 @@ fn main() -> Result<(), String> {
                 long_running = false;
             };
 
-            if let Some(releases) = releases {
+            if let Some(releases) = &releases {
                 query = query.releases(releases);
                 long_running = false;
             };
 
-            if let Some(users) = &users {
-                query = query.users(users.iter().map(|s| s.as_str()).collect());
+            if let Some(user_refs) = &user_refs {
+                query = query.users(user_refs);
                 long_running = false;
             };
 
@@ -506,10 +505,10 @@ fn main() -> Result<(), String> {
                 return Ok(());
             }
 
-            let result: Vec<Override> = match bodhi.query(query) {
-                Ok(overrides) => overrides,
-                Err(error) => return Err(error.to_string()),
-            };
+            let result: Vec<Override> = bodhi
+                .paginated_request(&query)
+                .await
+                .map_err(|error| error.to_string())?;
 
             pretty_outputs(&result, format)?;
 
@@ -517,8 +516,6 @@ fn main() -> Result<(), String> {
         },
         BodhiCommand::QueryUpdates {
             alias,
-            approved_before,
-            approved_since,
             bugs,
             builds,
             critpath,
@@ -544,31 +541,26 @@ fn main() -> Result<(), String> {
         } => {
             let format = format.unwrap_or(Format::Plain);
 
+            let aliases = alias.as_ref().map(|alias| vec![alias.as_str()]);
+            let build_refs: Option<Vec<&str>> = builds.as_ref().map(|bs| bs.iter().map(|b| b.as_str()).collect());
+            let pkg_refs: Option<Vec<&str>> = packages.as_ref().map(|ps| ps.iter().map(|p| p.as_str()).collect());
+            let user_refs: Option<Vec<&str>> = users.as_ref().map(|us| us.iter().map(|u| u.as_str()).collect());
+
             let mut query = UpdateQuery::new();
             let mut long_running = true;
 
-            if let Some(alias) = &alias {
-                query = query.aliases(vec![alias]);
+            if let Some(aliases) = &aliases {
+                query = query.aliases(aliases);
                 long_running = false;
             };
 
-            if let Some(approved_before) = &approved_before {
-                query = query.approved_before(approved_before);
-                long_running = false;
-            };
-
-            if let Some(approved_since) = &approved_since {
-                query = query.approved_since(approved_since);
-                long_running = false;
-            };
-
-            if let Some(bugs) = bugs {
+            if let Some(bugs) = &bugs {
                 query = query.bugs(bugs);
                 long_running = false;
             };
 
-            if let Some(builds) = &builds {
-                query = query.builds(builds.iter().map(|s| s.as_str()).collect());
+            if let Some(build_refs) = &build_refs {
+                query = query.builds(build_refs);
                 long_running = false;
             };
 
@@ -597,8 +589,8 @@ fn main() -> Result<(), String> {
                 long_running = false;
             };
 
-            if let Some(packages) = &packages {
-                query = query.packages(packages.iter().map(|s| s.as_str()).collect());
+            if let Some(pkg_refs) = &pkg_refs {
+                query = query.packages(pkg_refs);
                 long_running = false;
             };
 
@@ -617,7 +609,7 @@ fn main() -> Result<(), String> {
                 long_running = false;
             };
 
-            if let Some(releases) = releases {
+            if let Some(releases) = &releases {
                 query = query.releases(releases);
                 long_running = false;
             };
@@ -657,8 +649,8 @@ fn main() -> Result<(), String> {
                 long_running = false;
             };
 
-            if let Some(users) = &users {
-                query = query.users(users.iter().map(|s| s.as_str()).collect());
+            if let Some(user_refs) = &user_refs {
+                query = query.users(user_refs);
                 long_running = false;
             };
 
@@ -673,62 +665,60 @@ fn main() -> Result<(), String> {
                 return Ok(());
             }
 
-            let result: Vec<Update> = match bodhi.query(query) {
-                Ok(updates) => updates,
-                Err(error) => return Err(error.to_string()),
-            };
+            let result: Vec<Update> = bodhi
+                .paginated_request(&query)
+                .await
+                .map_err(|error| error.to_string())?;
 
             pretty_outputs(&result, format)?;
 
             Ok(())
         },
         BodhiCommand::ReleaseInfo { release, format } => {
-            let result: Option<Release> = match bodhi.query(ReleaseNameQuery::new(&release)) {
-                Ok(compose) => compose,
-                Err(_) => return Err(String::from("Failed to query releases.")),
-            };
+            let result: Release = bodhi
+                .request(&ReleaseNameQuery::new(&release))
+                .await
+                .map_err(|error| error.to_string())?;
 
-            pretty_output(
-                result.as_ref(),
-                &release,
-                "No release found with name",
-                format.unwrap_or(Format::Plain),
-            )?;
+            pretty_output(&result, format.unwrap_or(Format::Plain))?;
 
             Ok(())
         },
         BodhiCommand::ReleaseList { format } => {
-            let result: Vec<Release> = match bodhi.query(ReleaseQuery::new()) {
-                Ok(releases) => releases,
-                Err(error) => return Err(error.to_string()),
-            };
+            let result: Vec<Release> = bodhi
+                .paginated_request(&ReleaseQuery::new())
+                .await
+                .map_err(|error| error.to_string())?;
 
             pretty_outputs(&result, format.unwrap_or(Format::Plain))?;
 
             Ok(())
         },
         BodhiCommand::UpdateRequest { alias, request } => {
-            let update: Update = query_update(&bodhi, &alias)?;
+            let update: Update = query_update(&bodhi, &alias).await?;
             let editor = UpdateStatusRequester::from_update(&update, request);
 
-            let result: Update = match bodhi.edit(&editor) {
-                Ok(value) => value,
-                Err(error) => return Err(error.to_string()),
-            };
+            let result: Update = bodhi.request(&editor).await.map_err(|error| error.to_string())?;
 
+            println!("Update requested for {}.", request);
             println!("{}", result);
 
             Ok(())
         },
-        BodhiCommand::WaiveTests { alias, comment } => {
-            let update = query_update(&bodhi, &alias)?;
-            let editor = UpdateTestResultWaiver::from_update(&update, &comment);
+        BodhiCommand::WaiveTests { alias, comment, tests } => {
+            let update = query_update(&bodhi, &alias).await?;
 
-            let result: Update = match bodhi.edit(&editor) {
-                Ok(value) => value,
-                Err(error) => return Err(error.to_string()),
-            };
+            let test_refs: Option<Vec<&str>> = tests.as_ref().map(|ts| ts.iter().map(|t| t.as_str()).collect());
 
+            let mut editor = UpdateTestResultWaiver::from_update(&update, &comment);
+
+            if let Some(test_refs) = &test_refs {
+                editor = editor.tests(test_refs)
+            }
+
+            let result: Update = bodhi.request(&editor).await.map_err(|error| error.to_string())?;
+
+            println!("Tests waived.");
             println!("{}", result);
 
             Ok(())
